@@ -1,24 +1,7 @@
-use crate::{
-  config::Config,
-  error::{self, Error, Result},
-  file_stream::FileStream,
-  html,
-  input_path::InputPath,
-  redirect::redirect,
-};
+use crate::{common::*, file_stream::FileStream};
 use agora_lnd_client::lnrpc::invoice::InvoiceState;
-use hyper::{header, Body, Request, Response, StatusCode};
-use lexiclean::Lexiclean;
-use maud::{html, Markup};
+use maud::html;
 use percent_encoding::{AsciiSet, NON_ALPHANUMERIC};
-use snafu::{IntoError, ResultExt};
-use std::{
-  ffi::OsString,
-  fmt::Debug,
-  fs::{self, FileType},
-  io,
-  path::Path,
-};
 
 #[derive(Clone, Debug)]
 pub(crate) struct Files {
@@ -37,8 +20,8 @@ impl Files {
     }
   }
 
-  fn tail_to_path(&self, tail: &[&str]) -> Result<InputPath> {
-    self.base_directory.join_file_path(&tail.join(""))
+  fn file_path(&self, path: &str) -> Result<InputPath> {
+    self.base_directory.join_file_path(path)
   }
 
   fn check_path(&self, path: &InputPath) -> Result<()> {
@@ -94,7 +77,7 @@ impl Files {
     request: &Request<Body>,
     tail: &[&str],
   ) -> Result<Response<Body>> {
-    let file_path = self.tail_to_path(tail)?;
+    let file_path = self.file_path(&tail.join(""))?;
 
     for result in self.base_directory.iter_prefixes(tail) {
       let prefix = result?;
@@ -120,7 +103,7 @@ impl Files {
     if file_type.is_dir() {
       self.serve_dir(&file_path).await
     } else {
-      self.access_file(tail, &file_path).await
+      self.access_file(request, tail, &file_path).await
     }
   }
 
@@ -148,7 +131,29 @@ impl Files {
     Ok(entries)
   }
 
-  const ENCODE_CHARACTERS: AsciiSet = NON_ALPHANUMERIC.remove(b'/');
+  // Percent encode all unicode codepoints, even though
+  // they are allowed by the spec:
+  // https://url.spec.whatwg.org/#url-code-points
+  const ENCODE_CHARACTERS: AsciiSet = NON_ALPHANUMERIC
+    .remove(b'!')
+    .remove(b'$')
+    .remove(b'&')
+    .remove(b'\'')
+    .remove(b'(')
+    .remove(b')')
+    .remove(b'*')
+    .remove(b'+')
+    .remove(b',')
+    .remove(b'-')
+    .remove(b'.')
+    .remove(b'/')
+    .remove(b':')
+    .remove(b';')
+    .remove(b'=')
+    .remove(b'?')
+    .remove(b'@')
+    .remove(b'_')
+    .remove(b'~');
 
   fn render_index(dir: &InputPath) -> Result<Option<Markup>> {
     use pulldown_cmark::{html, Options, Parser};
@@ -190,7 +195,7 @@ impl Files {
             }
             @if file_type.is_file() && !config.paid() {
               a download href=(encoded) {
-                (Files::download_icon())
+                (Files::icon("download"))
               }
             }
           }
@@ -205,15 +210,20 @@ impl Files {
     Ok(html::wrap_body(body))
   }
 
-  fn download_icon() -> Markup {
+  fn icon(name: &str) -> Markup {
     html! {
       svg class="icon" {
-        use href="/static/feather-sprite.svg#download" {}
+        use href=(format!("/static/feather-sprite.svg#{}",name)) {}
       }
     }
   }
 
-  async fn access_file(&mut self, tail: &[&str], path: &InputPath) -> Result<Response<Body>> {
+  async fn access_file(
+    &mut self,
+    request: &Request<Body>,
+    tail: &[&str],
+    path: &InputPath,
+  ) -> Result<Response<Body>> {
     let config = self.config_for_dir(
       path
         .as_ref()
@@ -244,9 +254,9 @@ impl Files {
       .await
       .context(error::LndRpcStatus)?;
     redirect(format!(
-      "/invoice/{}/{}",
+      "{}?invoice={}",
+      request.uri().path(),
       hex::encode(invoice.r_hash),
-      file_path,
     ))
   }
 
@@ -263,6 +273,7 @@ impl Files {
   pub(crate) async fn serve_invoice(
     &mut self,
     request: &Request<Body>,
+    request_tail: &[&str],
     r_hash: [u8; 32],
   ) -> Result<Response<Body>> {
     let lnd_client = self.lnd_client.as_mut().ok_or_else(|| {
@@ -276,28 +287,79 @@ impl Files {
       .await
       .context(error::LndRpcStatus)?
       .ok_or_else(|| error::InvoiceNotFound { r_hash }.build())?;
+
+    let request_tail = request_tail.join("");
+    if request_tail != invoice.memo {
+      return Err(
+        error::InvoicePathMismatch {
+          invoice_tail: invoice.memo,
+          request_tail,
+          r_hash,
+        }
+        .build(),
+      );
+    }
+
     let value = invoice.value_msat();
     match invoice.state() {
       InvoiceState::Settled => {
-        let tail_from_invoice = invoice.memo.split_inclusive('/').collect::<Vec<&str>>();
-        let path = self.tail_to_path(&tail_from_invoice)?;
+        let path = self.file_path(&invoice.memo)?;
         Self::serve_file(&path).await
       }
       _ => {
         let qr_code_url = format!("/invoice/{}.svg", hex::encode(invoice.r_hash));
+        let filename = invoice.memo;
         Ok(html::wrap_body(html! {
           div class="invoice" {
             div class="label" {
-              (format!("Lightning Payment Request for {} to access ", value))
+              "Lightning Payment Request for " (value) " to access "
               span class="filename" {
-                  (invoice.memo)
+                  (filename)
               }
               ":"
             }
-            div class="payment-request" {
+            div class="payment-request"{
+              button class="clipboard-copy" onclick=(
+                format!("navigator.clipboard.writeText(\"{}\")", invoice.payment_request)
+              ) {
+                (Files::icon("clipboard"))
+              }
               (invoice.payment_request)
             }
-            img class="qr-code" alt="Lightning Network Invoice QR Code" src=(qr_code_url);
+
+            div class="links" {
+              a class="payment-link" href={"lightning:" (invoice.payment_request)} {
+                "Open invoice in wallet"
+              }
+              a class="reload-link" href=(request.uri()) {
+                "Access file"
+              }
+            }
+            img
+              class="qr-code"
+              alt="Lightning Network Invoice QR Code"
+              src=(qr_code_url)
+              width="400"
+              height="400";
+          }
+          div class="instructions" {
+            "To access "
+            span class="filename" {
+                (filename)
+            }
+            ":"
+            ol {
+              li {
+                "Pay the invoice for " (value) " above "
+                "with your Lightning Network wallet by "
+                "scanning the QR code, "
+                "copying the payment request string, or "
+                "clicking the \"Open invoice in wallet\" link."
+              }
+              li {
+                "Click the \"Access file\" link or reload the page."
+              }
+            }
           }
         }))
       }
