@@ -1,5 +1,4 @@
-use crate::error::{self, Error, Result};
-use hyper::{header, Body, Response};
+use crate::common::*;
 use rust_embed::RustEmbed;
 
 #[derive(RustEmbed)]
@@ -16,7 +15,7 @@ impl StaticAssets {
           builder = builder.header(header::CONTENT_TYPE, guess.essence_str());
         }
         builder
-          .body(bytes.into())
+          .body(bytes.data.into())
           .map_err(|error| Error::internal(format!("Failed to construct response: {}", error)))
       }
       None => Err(error::StaticAssetNotFound { uri_path: path }.build()),
