@@ -1,21 +1,20 @@
 use crate::{
-  environment::Environment,
-  error::Error,
-  server::Server,
-  test_utils::{assert_contains, assert_not_contains, test, test_with_environment, TestContext},
+  common::*,
+  server::TestContext,
+  test_utils::{
+    assert_contains, assert_not_contains, https_client, set_up_test_certificate, test,
+    test_with_arguments, test_with_environment,
+  },
 };
 use guard::guard_unwrap;
-use hyper::{header, StatusCode};
-use lexiclean::Lexiclean;
 use pretty_assertions::assert_eq;
 use reqwest::{redirect::Policy, Client, Url};
 use scraper::{ElementRef, Html, Selector};
-use std::{fs, net::TcpListener, path::Path, path::MAIN_SEPARATOR, str};
-use tokio::{
-  io::{AsyncReadExt, AsyncWriteExt},
-  net::TcpStream,
-};
+use std::{net::TcpListener, path::MAIN_SEPARATOR};
+use tokio::net::TcpStream;
 
+#[cfg(feature = "slow-tests")]
+mod browser_tests;
 #[cfg(feature = "slow-tests")]
 mod slow_tests;
 
@@ -71,7 +70,7 @@ fn configure_port() {
     "agora".into(),
     "--address=localhost".into(),
     "--directory=www".into(),
-    "--port".into(),
+    "--http-port".into(),
     free_port.to_string().into(),
   ];
   let www = environment.working_directory.join("www");
@@ -143,7 +142,7 @@ fn index_route_contains_title() {
 fn server_aborts_when_directory_does_not_exist() {
   let mut environment = Environment::test();
 
-  tokio::runtime::Builder::new_current_thread()
+  tokio::runtime::Builder::new_multi_thread()
     .enable_all()
     .build()
     .unwrap()
@@ -162,8 +161,7 @@ fn errors_in_request_handling_cause_500_status_codes() {
   use std::os::unix::fs::PermissionsExt;
 
   let stderr = test(|context| async move {
-    let file = context.files_directory().join("foo");
-    fs::write(&file, "").unwrap();
+    let file = context.write("foo", "");
     let mut permissions = file.metadata().unwrap().permissions();
     permissions.set_mode(0o000);
     fs::set_permissions(file, permissions).unwrap();
@@ -183,7 +181,7 @@ fn errors_in_request_handling_cause_500_status_codes() {
 #[test]
 fn listing_contains_file() {
   test(|context| async move {
-    fs::write(context.files_directory().join("some-test-file.txt"), "").unwrap();
+    context.write("some-test-file.txt", "");
     let haystack = html(context.base_url()).await.root_element().html();
     let needle = "some-test-file.txt";
     assert_contains(&haystack, needle);
@@ -193,8 +191,8 @@ fn listing_contains_file() {
 #[test]
 fn listing_contains_multiple_files() {
   test(|context| async move {
-    fs::write(context.files_directory().join("a.txt"), "").unwrap();
-    fs::write(context.files_directory().join("b.txt"), "").unwrap();
+    context.write("a.txt", "");
+    context.write("b.txt", "");
     let haystack = html(context.base_url()).await.root_element().html();
     assert_contains(&haystack, "a.txt");
     assert_contains(&haystack, "b.txt");
@@ -204,11 +202,11 @@ fn listing_contains_multiple_files() {
 #[test]
 fn listing_is_sorted_alphabetically() {
   test(|context| async move {
-    fs::write(context.files_directory().join("b"), "").unwrap();
-    fs::write(context.files_directory().join("c"), "").unwrap();
-    fs::write(context.files_directory().join("a"), "").unwrap();
+    context.write("b", "");
+    context.write("c", "");
+    context.write("a", "");
     let html = html(context.base_url()).await;
-    let haystack: Vec<&str> = css_select(&html, "a:not([download])")
+    let haystack: Vec<&str> = css_select(&html, ".listing a:not([download])")
       .into_iter()
       .map(|x| x.text())
       .flatten()
@@ -220,13 +218,9 @@ fn listing_is_sorted_alphabetically() {
 #[test]
 fn listed_files_can_be_played_in_browser() {
   test(|context| async move {
-    fs::write(
-      context.files_directory().join("some-test-file.txt"),
-      "contents",
-    )
-    .unwrap();
+    context.write("some-test-file.txt", "contents");
     let html = html(context.files_url()).await;
-    guard_unwrap!(let &[a] = css_select(&html, "a:not([download])").as_slice());
+    guard_unwrap!(let &[a] = css_select(&html, ".listing a:not([download])").as_slice());
     assert_eq!(a.inner_html(), "some-test-file.txt");
     let file_url = a.value().attr("href").unwrap();
     let file_url = context.files_url().join(file_url).unwrap();
@@ -238,11 +232,7 @@ fn listed_files_can_be_played_in_browser() {
 #[test]
 fn listed_files_have_download_links() {
   test(|context| async move {
-    fs::write(
-      context.files_directory().join("some-test-file.txt"),
-      "contents",
-    )
-    .unwrap();
+    context.write("some-test-file.txt", "contents");
     let html = html(context.files_url()).await;
     guard_unwrap!(let &[a] = css_select(&html, "a[download]").as_slice());
     assert_contains(&a.inner_html(), "download");
@@ -256,15 +246,9 @@ fn listed_files_have_download_links() {
 #[test]
 fn listed_files_have_percent_encoded_hrefs() {
   test(|context| async move {
-    fs::write(
-      context
-        .files_directory()
-        .join("filename with special chäracters"),
-      "",
-    )
-    .unwrap();
+    context.write("filename with special chäracters", "");
     let html = html(context.base_url()).await;
-    let links = css_select(&html, "a");
+    let links = css_select(&html, ".listing a");
     assert_eq!(links.len(), 2);
     for link in links {
       assert_eq!(
@@ -358,7 +342,7 @@ fn configure_source_directory() {
   environment.arguments = vec![
     "agora".into(),
     "--address=localhost".into(),
-    "--port=0".into(),
+    "--http-port=0".into(),
     "--directory=src".into(),
   ];
 
@@ -412,7 +396,7 @@ fn downloaded_files_are_streamed() {
 #[test]
 fn downloaded_files_have_correct_content_type() {
   test(|context| async move {
-    fs::write(context.files_directory().join("foo.mp4"), "hello").unwrap();
+    context.write("foo.mp4", "hello");
 
     let response = get(&context.files_url().join("foo.mp4").unwrap()).await;
 
@@ -426,7 +410,7 @@ fn downloaded_files_have_correct_content_type() {
 #[test]
 fn unknown_files_have_no_content_type() {
   test(|context| async move {
-    fs::write(context.files_directory().join("foo"), "hello").unwrap();
+    context.write("foo", "hello");
 
     let response = get(&context.files_url().join("foo").unwrap()).await;
 
@@ -437,7 +421,7 @@ fn unknown_files_have_no_content_type() {
 #[test]
 fn filenames_with_spaces() {
   test(|context| async move {
-    fs::write(context.files_directory().join("foo bar"), "hello").unwrap();
+    context.write("foo bar", "hello");
 
     let response = text(&context.files_url().join("foo%20bar").unwrap()).await;
 
@@ -448,17 +432,16 @@ fn filenames_with_spaces() {
 #[test]
 fn subdirectories_appear_in_listings() {
   test(|context| async move {
-    fs::create_dir(context.files_directory().join("foo")).unwrap();
-    fs::write(context.files_directory().join("foo/bar.txt"), "hello").unwrap();
+    context.write("foo/bar.txt", "hello");
     let root_listing = html(context.files_url()).await;
-    guard_unwrap!(let &[a] = css_select(&root_listing, "a").as_slice());
+    guard_unwrap!(let &[a] = css_select(&root_listing, ".listing a").as_slice());
     assert_eq!(a.inner_html(), "foo/");
     let subdir_url = context
       .files_url()
       .join(a.value().attr("href").unwrap())
       .unwrap();
     let subdir_listing = html(&subdir_url).await;
-    guard_unwrap!(let &[a] = css_select(&subdir_listing, "a:not([download])").as_slice());
+    guard_unwrap!(let &[a] = css_select(&subdir_listing, ".listing a:not([download])").as_slice());
     assert_eq!(a.inner_html(), "bar.txt");
     let file_url = subdir_url.join(a.value().attr("href").unwrap()).unwrap();
     assert_eq!(text(&file_url).await, "hello");
@@ -477,10 +460,9 @@ fn no_trailing_slash_redirects_to_trailing_slash() {
 #[test]
 fn redirects_correctly_for_two_layers_of_subdirectories() {
   test(|context| async move {
-    fs::create_dir_all(context.files_directory().join("foo/bar")).unwrap();
-    fs::write(context.files_directory().join("foo/bar/baz.txt"), "").unwrap();
+    context.write("foo/bar/baz.txt", "");
     let listing = html(&context.files_url().join("foo/bar").unwrap()).await;
-    guard_unwrap!(let &[a] = css_select(&listing, "a:not([download])").as_slice());
+    guard_unwrap!(let &[a] = css_select(&listing, ".listing a:not([download])").as_slice());
     assert_eq!(a.inner_html(), "baz.txt")
   });
 }
@@ -509,7 +491,7 @@ fn file_errors_are_associated_with_file_path() {
 #[test]
 fn requesting_files_with_trailing_slash_redirects() {
   test(|context| async move {
-    fs::write(context.files_directory().join("foo"), "").unwrap();
+    context.write("foo", "");
     let response = reqwest::get(context.files_url().join("foo/").unwrap())
       .await
       .unwrap();
@@ -535,7 +517,7 @@ fn listings_are_not_cached() {
 #[test]
 fn files_are_not_cached() {
   test(|context| async move {
-    fs::write(context.files_directory().join("foo"), "bar").unwrap();
+    context.write("foo", "bar");
     let response = reqwest::get(context.files_url().join("foo").unwrap())
       .await
       .unwrap();
@@ -570,7 +552,7 @@ fn symlink(contents: impl AsRef<Path>, link: impl AsRef<Path>) {
 #[test]
 fn allow_file_downloads_via_local_symlinks() {
   test(|context| async move {
-    fs::write(&context.files_directory().join("file"), "contents").unwrap();
+    context.write("file", "contents");
     symlink("file", context.files_directory().join("link"));
     let response = reqwest::get(context.files_url().join("link").unwrap())
       .await
@@ -582,7 +564,7 @@ fn allow_file_downloads_via_local_symlinks() {
 #[test]
 fn disallow_file_downloads_via_escaping_symlinks() {
   let stderr = test(|context| async move {
-    fs::write(&context.files_directory().join("../file"), "contents").unwrap();
+    context.write("../file", "contents");
     symlink("../file", context.files_directory().join("link"));
     let response = reqwest::get(context.files_url().join("link").unwrap())
       .await
@@ -601,9 +583,9 @@ fn disallow_file_downloads_via_escaping_symlinks() {
 #[test]
 fn disallow_file_downloads_via_absolute_escaping_symlinks() {
   let stderr = test(|context| async move {
-    let file = context.files_directory().join("../file").lexiclean();
+    let file = context.write("../file", "contents");
+    let file = file.lexiclean();
     assert!(file.is_absolute());
-    fs::write(&file, "contents").unwrap();
     symlink(file, context.files_directory().join("link"));
     let response = reqwest::get(context.files_url().join("link").unwrap())
       .await
@@ -622,10 +604,8 @@ fn disallow_file_downloads_via_absolute_escaping_symlinks() {
 #[test]
 fn allow_file_downloads_via_local_intermediate_symlinks() {
   test(|context| async move {
-    let dir = context.files_directory().join("dir");
-    fs::create_dir(&dir).unwrap();
+    context.write("dir/file", "contents");
     symlink("dir", context.files_directory().join("link"));
-    fs::write(dir.join("file"), "contents").unwrap();
     let response = reqwest::get(context.files_url().join("link/file").unwrap())
       .await
       .unwrap();
@@ -636,10 +616,8 @@ fn allow_file_downloads_via_local_intermediate_symlinks() {
 #[test]
 fn disallow_file_downloads_via_escaping_intermediate_symlinks() {
   let stderr = test(|context| async move {
-    let dir = context.files_directory().join("../dir");
-    fs::create_dir(&dir).unwrap();
+    context.write("../dir/file", "contents");
     symlink("../dir", context.files_directory().join("link"));
-    fs::write(dir.join("file"), "contents").unwrap();
     let response = reqwest::get(context.files_url().join("link/file").unwrap())
       .await
       .unwrap();
@@ -725,11 +703,10 @@ fn disallow_listing_directories_via_intermediate_escaping_symlinks() {
 #[test]
 fn show_local_symlinks_in_listings() {
   test(|context| async move {
-    let file = context.files_directory().join("file");
-    fs::write(&file, "").unwrap();
-    symlink(file, context.files_directory().join("link"));
+    context.write("file", "");
+    symlink("file", context.files_directory().join("link"));
     let html = html(context.files_url()).await;
-    guard_unwrap!(let &[a, b] = css_select(&html, "a:not([download])").as_slice());
+    guard_unwrap!(let &[a, b] = css_select(&html, ".listing a:not([download])").as_slice());
     assert_eq!(a.inner_html(), "file");
     assert_eq!(b.inner_html(), "link");
   });
@@ -738,11 +715,11 @@ fn show_local_symlinks_in_listings() {
 #[test]
 fn remove_escaping_symlinks_from_listings() {
   test(|context| async move {
-    fs::write(&context.files_directory().join("../escaping"), "").unwrap();
-    fs::write(&context.files_directory().join("local"), "").unwrap();
+    context.write("../escaping", "");
+    context.write("local", "");
     symlink("../escaping", context.files_directory().join("link"));
     let html = html(context.files_url()).await;
-    guard_unwrap!(let &[a] = css_select(&html, "a:not([download])").as_slice());
+    guard_unwrap!(let &[a] = css_select(&html, ".listing a:not([download])").as_slice());
     assert_eq!(a.inner_html(), "local");
   });
 }
@@ -780,7 +757,7 @@ fn missing_asset_not_found() {
 #[test]
 fn listing_does_not_contain_hidden_file() {
   test(|context| async move {
-    fs::write(context.files_directory().join(".some-test-file.txt"), "").unwrap();
+    context.write(".some-test-file.txt", "");
     let haystack = html(context.base_url()).await.root_element().html();
     let needle = ".some-test-file.txt";
     assert_not_contains(&haystack, needle);
@@ -790,7 +767,7 @@ fn listing_does_not_contain_hidden_file() {
 #[test]
 fn return_404_for_hidden_files() {
   test(|context| async move {
-    fs::write(context.files_directory().join(".foo.txt"), "").unwrap();
+    context.write(".foo.txt", "");
     assert_eq!(
       reqwest::get(context.files_url().join(".foo.txt").unwrap())
         .await
@@ -816,9 +793,7 @@ fn return_404_for_hidden_directories() {
 #[test]
 fn return_404_for_files_in_hidden_directories() {
   test(|context| async move {
-    let dir = context.files_directory().join(".dir");
-    fs::create_dir(&dir).unwrap();
-    fs::write(dir.join("foo.txt"), "hello").unwrap();
+    context.write(".dir/foo.txt", "hello");
     let response = reqwest::get(context.files_url().join(".dir/foo.txt").unwrap())
       .await
       .unwrap();
@@ -829,8 +804,8 @@ fn return_404_for_files_in_hidden_directories() {
 #[test]
 fn requesting_paid_file_with_no_lnd_returns_internal_error() {
   let stderr = test(|context| async move {
-    fs::write(context.files_directory().join(".agora.yaml"), "paid: true").unwrap();
-    fs::write(context.files_directory().join("foo"), "precious content").unwrap();
+    context.write(".agora.yaml", "paid: true");
+    context.write("foo", "precious content");
     let status = reqwest::get(context.files_url().join("foo").unwrap())
       .await
       .unwrap()
@@ -847,24 +822,10 @@ fn requesting_paid_file_with_no_lnd_returns_internal_error() {
   );
 }
 
-#[test]
-fn errors_contain_backtraces() {
-  let stderr = test(|context| async move {
-    fs::write(context.files_directory().join(".hidden"), "").unwrap();
-    let status = reqwest::get(context.files_url().join(".hidden").unwrap())
-      .await
-      .unwrap()
-      .status();
-    assert_eq!(status, StatusCode::NOT_FOUND);
-  });
-
-  assert_contains(&stderr, "agora::files::Files::check_path");
-}
-
 #[test]
 fn displays_index_markdown_files_as_html() {
   test(|context| async move {
-    fs::write(context.files_directory().join(".index.md"), "# test header").unwrap();
+    context.write(".index.md", "# test header");
     let html = html(context.files_url()).await;
     guard_unwrap!(let &[index_header] = css_select(&html, "h1").as_slice());
     assert_eq!(index_header.inner_html(), "test header");
@@ -894,12 +855,8 @@ fn returns_error_if_index_is_unusable() {
 #[test]
 fn ignores_access_config_outside_of_base_directory() {
   test(|context| async move {
-    fs::write(
-      context.temp_directory().join(".agora.yaml"),
-      "{paid: true, base-price: 1000 sat}",
-    )
-    .unwrap();
-    fs::write(context.files_directory().join("foo"), "foo").unwrap();
+    context.write("../.agora.yaml", "{paid: true, base-price: 1000 sat}");
+    context.write("foo", "foo");
     let body = text(&context.files_url().join("foo").unwrap()).await;
     assert_eq!(body, "foo");
   });
@@ -909,15 +866,203 @@ fn ignores_access_config_outside_of_base_directory() {
 fn paid_files_dont_have_download_button() {
   #![allow(clippy::unused_unit)]
   test(|context| async move {
-    fs::write(
-      context.files_directory().join(".agora.yaml"),
-      "{paid: true, base-price: 1000 sat}",
-    )
-    .unwrap();
-    fs::write(context.files_directory().join("foo"), "foo").unwrap();
+    context.write(".agora.yaml", "{paid: true, base-price: 1000 sat}");
+    context.write("foo", "foo");
     let html = html(context.files_url()).await;
-    guard_unwrap!(let &[] = css_select(&html, "a[download]").as_slice());
-    guard_unwrap!(let &[link] = css_select(&html, "a:not([download])").as_slice());
+    guard_unwrap!(let &[] = css_select(&html, ".listing a[download]").as_slice());
+    guard_unwrap!(let &[link] = css_select(&html, ".listing a:not([download])").as_slice());
     assert_eq!(link.inner_html(), "foo");
   });
 }
+
+#[test]
+fn filenames_with_percent_encoded_characters() {
+  test(|context| async move {
+    context.write("=", "contents");
+    let contents = text(&context.files_url().join("%3D").unwrap()).await;
+    assert_eq!(contents, "contents");
+    let contents = text(&context.files_url().join("=").unwrap()).await;
+    assert_eq!(contents, "contents");
+  });
+}
+
+#[test]
+fn filenames_with_percent_encoding() {
+  test(|context| async move {
+    context.write("foo%20bar", "contents");
+    let contents = text(&context.files_url().join("foo%2520bar").unwrap()).await;
+    assert_eq!(contents, "contents");
+  });
+}
+
+#[test]
+fn filenames_with_invalid_percent_encoding() {
+  test(|context| async move {
+    context.write("%80", "contents");
+    let contents = text(&context.files_url().join("%2580").unwrap()).await;
+    assert_eq!(contents, "contents");
+  });
+}
+
+#[test]
+fn space_is_percent_encoded() {
+  test(|context| async move {
+    context.write("foo bar", "contents");
+    let html = html(context.files_url()).await;
+    guard_unwrap!(let &[a] = css_select(&html, ".listing a:not([download])").as_slice());
+    assert_eq!(a.value().attr("href").unwrap(), "foo%20bar");
+  });
+}
+
+#[test]
+fn doesnt_percent_encode_allowed_ascii_characters() {
+  test(|context| async move {
+    let allowed_ascii_characters = if cfg!(windows) {
+      "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz!$&'()+,-.;=@_~"
+    } else {
+      "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz!$&'()*+,-.:;=?@_~"
+    };
+    context.write(allowed_ascii_characters, "contents");
+    let html = html(context.files_url()).await;
+    guard_unwrap!(let &[a] = css_select(&html, ".listing a:not([download])").as_slice());
+    assert_eq!(a.value().attr("href").unwrap(), allowed_ascii_characters);
+  });
+}
+
+#[test]
+fn percent_encodes_unicode() {
+  test(|context| async move {
+    context.write("Å", "contents");
+    let html = html(context.files_url()).await;
+    guard_unwrap!(let &[a] = css_select(&html, ".listing a:not([download])").as_slice());
+    assert_eq!(a.value().attr("href").unwrap(), "%C3%85");
+  });
+}
+
+#[test]
+fn serves_https_requests_with_cert_from_cache_directory() {
+  let (certificate_cache, root_certificate) = set_up_test_certificate();
+
+  test_with_arguments(
+    &[
+      "--acme-cache-directory",
+      certificate_cache.path().to_str().unwrap(),
+      "--https-port=0",
+      "--acme-domain=localhost",
+    ],
+    |context| async move {
+      context.write("file", "encrypted content");
+      let client = https_client(&context, root_certificate).await;
+      let response = client
+        .get(context.https_files_url().join("file").unwrap())
+        .send()
+        .await
+        .unwrap();
+      let body = response.text().await.unwrap();
+      assert_eq!(body, "encrypted content");
+    },
+  );
+}
+
+#[test]
+fn creates_cert_cache_directory_if_it_doesnt_exist() {
+  test_with_arguments(
+    &[
+      "--acme-cache-directory",
+      "cache-directory",
+      "--https-port=0",
+      "--acme-domain=localhost",
+    ],
+    |context| async move {
+      let cache_directory = context.working_directory().join("cache-directory");
+      for _ in 0..100 {
+        if cache_directory.is_dir() {
+          return;
+        }
+        tokio::time::sleep(Duration::from_millis(100)).await;
+      }
+      panic!("Cache directory not created after ten seconds");
+    },
+  );
+}
+
+#[test]
+fn redirects_requests_from_port_80_to_443() {
+  let (certificate_cache, root_certificate) = set_up_test_certificate();
+
+  test_with_arguments(
+    &[
+      "--acme-cache-directory",
+      certificate_cache.path().to_str().unwrap(),
+      "--https-port=0",
+      "--https-redirect-port=0",
+      "--acme-domain=localhost",
+    ],
+    |context| async move {
+      context.write("file", "encrypted content");
+      let client = https_client(&context, root_certificate).await;
+      let response = client
+        .get(format!(
+          "http://localhost:{}/files/file",
+          context.https_redirect_port()
+        ))
+        .send()
+        .await
+        .unwrap();
+      assert!(response.url().to_string().starts_with("https:"));
+      let body = response.text().await.unwrap();
+      assert_eq!(body, "encrypted content");
+    },
+  );
+}
+
+#[test]
+fn favicon_is_served_at_favicon_ico() {
+  test(|context| async move {
+    let response = get(&context.base_url().join("favicon.ico").unwrap()).await;
+    assert_eq!(
+      response.headers().get(header::CONTENT_TYPE).unwrap(),
+      "image/x-icon"
+    );
+  });
+}
+
+#[test]
+fn apple_touch_icon_is_served_under_root() {
+  test(|context| async move {
+    let response = get(&context.base_url().join("apple-touch-icon.png").unwrap()).await;
+    assert_eq!(
+      response.headers().get(header::CONTENT_TYPE).unwrap(),
+      "image/png"
+    );
+  });
+}
+
+#[test]
+fn bugfix_symlink_with_relative_base_directory() {
+  let mut environment = Environment::test();
+
+  let www = environment.working_directory.join("www");
+  std::fs::create_dir(&www).unwrap();
+
+  let working_directory = environment.working_directory.join("working_directory");
+  std::fs::create_dir(&working_directory).unwrap();
+
+  environment.working_directory = working_directory;
+
+  environment.arguments = vec![
+    "agora".into(),
+    "--address=localhost".into(),
+    "--http-port=0".into(),
+    "--directory=../www".into(),
+  ];
+
+  test_with_environment(&mut environment, |context| async move {
+    context.write("file", "precious content");
+    symlink("file", context.files_directory().join("link"));
+    let content = text(&context.files_url().join("file").unwrap()).await;
+    assert_eq!(content, "precious content");
+    let link = text(&context.files_url().join("link").unwrap()).await;
+    assert_eq!(link, "precious content");
+  });
+}
